#@ File (label = "image" ) imgpath
#@ Float (label = "LoG sigma 1-2", value = 1) sigma
#@ String (label = "Register stack", choices = {"SIFT", "None"}) reg_method
#@ Float (label = "find maxima prominence 0.5-1", value = 0.5) prominence
#@ String (label = "Threshold method", choices = {"Otsu", "Li"}) thr_method
#@ Float (label = "Threshold factor", value = 1) thrfactor
#@ Integer (label = "Radius dilation spots", value = 15) spot_dilation
#@ Integer (label = "Min area mask", value = 50) min_area_mask
#@ Float  (label = "threshold mask", value = 11) thr_mask
#@ Output Float densitymask
#@ Output Float areamask
#@ Output Float densityconvex
#@ Output Float areaconvex

import ij.*
import ij.plugin.ZProjector as ZProjector
import ij.plugin.frame.RoiManager as RoiManager
import ij.gui.PointRoi as PointRoi
import imagescience.feature.Laplacian as Laplacian
import imagescience.image.Image as Image
import com.opencsv.CSVWriter
import java.util.Collections;


// Antonio Politi
// apoliti@mpibpc.mpg.de MPIBPC April 2021
// Script to find small spots, e.g. NPCs
// register image stack 
// filter by LoG
// perform spot finding (find maxima) on sum-projected laplacian image
// Select spots according to threshold in laplace and raw image

// TODO: Check for packages

 // Initial image is already single cell

main();

def main() {
	
	ver = IJ.getVersion()
	IJ.run("Close All");
    RoiManager rm;
	rm = new RoiManager();
	rm = RoiManager.getInstance()
	if (rm.getCount() > 0) {
		rm.reset();
	}


	imp = IJ.openImage(imgpath.toString());
	imp.show();
	try {
		IJ.run(imp, "signed16bit to unsigned", "");
	} catch (Exception e) {
		IJ.showMessage("Please enable the liveim-tools update site! [Help > Update > Manage update site > [x] liveim-tools]")
	}
	imp.setTitle("raw");
	cal = imp.getCalibration();
	pixelSize = [cal.pixelWidth, cal.pixelHeight];
// Register the stack
	imp_aligned = register(imp, reg_method);
	imp_aligned.setTitle("registered");


// Compute mask of dots


// LoG filtering
	imp_laplace = laplace(imp_aligned, sigma);
	imp_laplace.setTitle("registered_laplace")

	imp_laplace_Zproj = ZProjector.run(imp_laplace, "sum");
	imp_laplace_Zproj.show();

// Smoothed sum image
	imp_Zproj = ZProjector.run(imp_aligned, "sum");
	IJ.run(imp_Zproj, "Smooth", "");
	imp_Zproj.show();

// Find spots 
	IJ.run(imp_laplace_Zproj, "Find Maxima...", "prominence=" + prominence + " output=[Point Selection]");
	points = imp_laplace_Zproj.getRoi();
	points.setName("spots");

// Filter spots
	ips = [imp_laplace_Zproj.getProcessor(), imp_Zproj.getProcessor()];
	ips[0].setAutoThreshold(thr_method)
	ips[1].setAutoThreshold(thr_method)
	thr = [ips[0].getMaxThreshold() * thrfactor, ips[1].getMaxThreshold() * thrfactor];

	points_filter = filter_maxima(points, ips, thr);
	points_filter.setName("spots_filter");
	rois_spots = [points, points_filter];

// Create mask
	roiMask = createMaskFromPores(imp, rm, thr_mask, spot_dilation, min_area_mask);

// Filter pores using the mask
	rois_spots = maskPores(rois_spots, roiMask, rm, imp);

	// Add rois
	for (roi in rois_spots) {
		rm.addRoi(roi);
	}

	// Create output
	rm.runCommand("save", imgpath.toString() + ".zip");
	nrEl = rm.getCount();
	imp.setRoi(rois_spots[nrEl-1]);
	areaconvex = imp.getStatistics().area;
	if (roiMask[0]) {
		imp.setRoi(rois_spots[nrEl-3]);
		areamask = imp.getStatistics().area;
	} else {
		areamask = 1;
	}

	IJ.log("" + imp.getStatistics().area);
	File outputfile = new File(imgpath.toString() + ".csv");
	nrspots = savePointsCsv(rois_spots[nrEl-2], ips, pixelSize, areaconvex, areamask, outputfile);
	densityconvex = nrspots/areaconvex;
	densitymask = nrspots/areamask;
	imp.show();

}

def maskPores(spots, mask, rm, imp){
	if (!mask[0]) {
		IJ.log("no Nuclei Roi found use the spots");
		imp.setRoi(spots[1]);
		IJ.run(imp, "Convex Hull", "");
		roiConvex = imp.getRoi();
		roiConvex.setName("convex_area");
		spots.add(roiConvex);
		return spots;

	} else {

		spots.add(mask[1]);

		for (roi in spots){
			rm.addRoi(roi);
		}
		rm.setSelectedIndexes(new int[]{1, 2});
		rm.runCommand(imp, "AND");
		points_filter_masked = imp.getRoi();
		points_filter_masked.setName("spots_filter_masked");
		IJ.run(imp, "Convex Hull", "");
		roiConvex = imp.getRoi();
		roiConvex.setName("Convex_area");
		spots.add(points_filter_masked);
		spots.add(roiConvex);
		rm.reset();
	}
	return spots;
}










// A second step is to perform an and operation with mask of nuclei this is a separate script

def createMaskFromPores(imp, rm, thr, opening_radius, nuclei_area) {
	IJ.run(imp, "Morphological Filters", "operation=Dilation element=Disk radius=" + opening_radius);
	imp_dilated = IJ.getImage();
	maxValue = imp_dilated.getProcessor().getMax();
	imp_dilated.getProcessor().setThreshold(thr, maxValue, 1);
	IJ.run(imp_dilated, "Convert to Mask", "");
	IJ.run(imp_dilated, "Analyze Particles...", "size="+nuclei_area+"-Infinity display include add");
	if (rm.getCount() == 0)
		return [false, 0];
	// local variable
	def rois = rm.getRoisAsArray();
	rm.reset();
	List <Double> area = new ArrayList<>();
	for (roi in rois){
		area.add(roi.getStatistics().area);
	}
	int ind_area = area.indexOf(Collections.max(area));
	rois[ind_area].setName("mask");
	return [true, rois[ind_area]];
}



def register(imp, method) {
	// return a registered image or the same image depending on the method
	if (method == "SIFT"){
		imp.show();
		IJ.run("Linear Stack Alignment with SIFT", "initial_gaussian_blur=1.60 steps_per_scale_octave=3 minimum_image_size=64 maximum_image_size=1024 feature_descriptor_size=4 feature_descriptor_orientation_bins=8 closest/next_closest_ratio=0.92 maximal_alignment_error=25 inlier_ratio=0.05 expected_transformation=Rigid interpolate");
		imp_aligned = IJ.getImage();
		return imp_aligned;
	} else {
		imp_aligned = imp.duplicate();
		return imp;
	}
}

def laplace(imp, sigma){
	// Compute LoG of image
	
	//IJ.run("FeatureJ Laplacian", "compute smoothing=1"); // 1 or 1.5 do not pick something too high
	//lap = Class.forName("imagescience.feature.Laplacian").newInstance()

	lap = new Laplacian();
	// This is a 3D laplacian!
	imp_laplace = lap.run(Image.wrap(imp), sigma);
	imp_laplace = imp_laplace.imageplus();
	IJ.run(imp_laplace, "Multiply...", "value=-1 stack");
	return imp_laplace
}

def filter_maxima(points, ip, thr) {
	// filter maxima according to some intensity thresholds
	PTs = points.getContainedPoints();
	points_clean = new PointRoi();
	def pixValue = [0, 0];
	for (int ipt = 0; ipt < PTs.size(); ipt++){
		pixValue[0] = ip[0].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
		pixValue[1] = ip[1].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
	
		if ( (pixValue[0] > thr[0]) && (pixValue[1] > thr[1])) {
			points_clean.addPoint(PTs[ipt].getX(), PTs[ipt].getY());
		}
	
	}
	return points_clean
}

def savePointsCsv(points, ip, pixelSize, area_convex, area_mask, outfile) {

	PTs = points.getContainedPoints();
	final List<String[]> dataout = new ArrayList<>();

	def pixValue = [0, 0];
	for (int ipt = 0; ipt < PTs.size(); ipt++){
		pixValue[0] = ip[0].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
		pixValue[1] = ip[1].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
		dataout.add(new String[]{ "" + PTs[ipt].getX() , "" + PTs[ipt].getY() ,
				"" + PTs[ipt].getX()*pixelSize[0] , "" + PTs[ipt].getY()*pixelSize[1],
				"" + pixValue[0], "" + pixValue[1],  "" +  area_convex, "" +  area_mask,
				"" +  PTs.size()/area_convex, "" +  PTs.size()/area_mask});
	}
	header = new String[]{"X_px", "Y_px", "X_um", "Y_um", "val_laplace",
			"val_raw", "area_convex_um2", "area_mask_um2", "density_convex_um2", "density_mask_um2"};
	writeDataAtOnce(outfile, header, dataout, false);
	return PTs.size();
}



def writeDataAtOnce(outfile, header,   data,  append)  {
	// first create file object for file placed at location
	// specified by filepath
	//File file = new File(filePath);
	try {
		if (!outfile.exists())
			append = false;
		FileWriter outputfile = new FileWriter(outfile, append);
		// create CSVWriter object filewriter object as parameter
		CSVWriter writer = new CSVWriter(outputfile);
		if (!append) {
			writer.writeNext(header);
		}
		writer.writeAll(data);
		writer.close();
	}
	catch (IOException e) {
		e.printStackTrace();
	}
}

def checkPackages() {
	try {
		Class.forName("imagescience.image.Image");
		return true;
	} catch (Exception e) {
		IJ.showMessage("Please enable the ImageScience update site! [Help > Update > Manage update site > [x] ImageScience] ")
		return false;
	}
}